from shadow4.beamline.s4_beamline import S4Beamline

beamline = S4Beamline()

# electron beam
from shadow4.sources.s4_electron_beam import S4ElectronBeam

electron_beam = S4ElectronBeam(energy_in_GeV=6, energy_spread=0.001, current=0.2)
electron_beam.set_sigmas_all(sigma_x=3.03e-05, sigma_y=3.6e-06, sigma_xp=4.4e-06, sigma_yp=1.46e-06)

# magnetic structure
from shadow4.sources.undulator.s4_undulator_gaussian import S4UndulatorGaussian

source = S4UndulatorGaussian(
    period_length=0.027,  # syned Undulator parameter (length in m)
    number_of_periods=74.0,  # syned Undulator parameter
    photon_energy=9000,  # Photon energy (in eV)
    delta_e=0.0,  # Photon energy width (in eV)
    ng_e=100,  # Photon energy scan number of points
    flag_emittance=1,  # when sampling rays: Use emittance (0=No, 1=Yes)
    flag_energy_spread=0,  # when sampling rays: Use e- energy spread (0=No, 1=Yes)
    harmonic_number=1,  # harmonic number
    flag_autoset_flux_central_cone=0,  # value to set the flux peak
    flux_central_cone=10000000000.0,  # value to set the flux peak
)

# light source
from shadow4.sources.undulator.s4_undulator_gaussian_light_source import S4UndulatorGaussianLightSource

light_source = S4UndulatorGaussianLightSource(name='GaussianUndulator', electron_beam=electron_beam,
                                              magnetic_structure=source, nrays=50000, seed=5676561)
beam = light_source.get_beam()

beamline.set_light_source(light_source)

# optical element number XX
from syned.beamline.shape import Circle

boundary_shape = Circle(radius=0.00022, x_center=0, y_center=0)
from shadow4.beamline.optical_elements.refractors.s4_crl import S4CRL

optical_element = S4CRL(name='Compound Refractive Lens',
                        n_lens=4,
                        piling_thickness=0.002,  # syned stuff
                        boundary_shape=boundary_shape,  # syned stuff, replaces "diameter" in the shadow3 append_lens
                        material='Be',  # the material for ri_calculation_mode > 1
                        density=1.848,  # the density for ri_calculation_mode > 1
                        thickness=9.999999999999999e-06,
                        # syned stuff, lens thickness [m] (distance between the two interfaces at the center of the lenses)
                        surface_shape=2,  # now: 0=plane, 1=sphere, 2=parabola, 3=conic coefficients
                        # (in shadow3: 1=sphere 4=paraboloid, 5=plane)
                        convex_to_the_beam=0,
                        # for surface_shape: convexity of the first interface exposed to the beam 0=No, 1=Yes
                        cylinder_angle=0,  # for surface_shape: 0=not cylindricaL, 1=meridional 2=sagittal
                        ri_calculation_mode=2,  # source of refraction indices and absorption coefficients
                        # 0=User, 1=prerefl file, 2=xraylib, 3=dabax
                        prerefl_file='/home/srio/Oasys/Be.dat',
                        # for ri_calculation_mode=0: file name (from prerefl) to get the refraction index.
                        refraction_index=1,  # for ri_calculation_mode=1: n (real)
                        attenuation_coefficient=0,  # for ri_calculation_mode=1: mu in cm^-1 (real)
                        dabax=None,  # the pointer to dabax library
                        radius=5e-05,
                        # for surface_shape=(1,2): lens radius [m] (for spherical, or radius at the tip for paraboloid)
                        conic_coefficients1=None,
                        # for surface_shape = 3: the conic coefficients of the single lens interface 1
                        conic_coefficients2=None,
                        # for surface_shape = 3: the conic coefficients of the single lens interface 2
                        )

import numpy
from syned.beamline.element_coordinates import ElementCoordinates

coordinates = ElementCoordinates(p=117, q=1.50255, angle_radial=0, angle_azimuthal=0, angle_radial_out=3.141592654)
movements = None
from shadow4.beamline.optical_elements.refractors.s4_crl import S4CRLElement

beamline_element = S4CRLElement(optical_element=optical_element, coordinates=coordinates, movements=movements,
                                input_beam=beam)

beam, mirr = beamline_element.trace_beam()

beamline.append_beamline_element(beamline_element)

# optical element number XX
from hybrid_methods.coherence.hybrid_screen import StdIOHybridListener
from shadow4_advanced.hybrid.s4_hybrid_screen import S4HybridBeam, S4HybridOE, HybridInputParameters, \
    S4HybridScreenElement
from shadow4_advanced.hybrid.s4_hybrid_screen import S4HybridScreen

calculation_type = 5
hybrid_screen = S4HybridScreen(calculation_type)

additional_parameters = {}
additional_parameters["crl_error_profiles"] = ['/home/srio/Oasys/dabam2d-0022.h5', '/home/srio/Oasys/dabam2d-0023.h5', '/home/srio/Oasys/dabam2d-0024.h5',
                                               '/home/srio/Oasys/dabam2d-0025.h5']
additional_parameters["crl_material"] = 'Be'
additional_parameters["crl_scaling_factor"] = 1
additional_parameters["crl_coords_to_m"] = 1
additional_parameters["crl_thickness_to_m"] = 1
hybrid_beamline_element = beamline.get_beamline_element_at(-1)
input_parameters = HybridInputParameters(listener=StdIOHybridListener(),
                                         beam=S4HybridBeam(beam=beam),
                                         optical_element=S4HybridOE(optical_element=hybrid_beamline_element),
                                         diffraction_plane=2,
                                         propagation_type=0,
                                         n_bins_x=14,
                                         n_bins_z=14,
                                         n_peaks=10,
                                         fft_n_pts=100000,
                                         analyze_geometry=False,
                                         random_seed=None,
                                         **additional_parameters)

beamline_element = S4HybridScreenElement(hybrid_screen=hybrid_screen, hybrid_input_parameters=input_parameters)

beam, mirr, _ = beamline_element.trace_beam()

beamline.append_beamline_element(beamline_element)

# test plot
if True:
    from srxraylib.plot.gol import plot_scatter

    plot_scatter(beam.get_photon_energy_eV(nolost=1), beam.get_column(23, nolost=1), title='(Intensity,Photon Energy)',
                 plot_histograms=0)
    plot_scatter(1e6 * beam.get_column(1, nolost=1), 1e6 * beam.get_column(3, nolost=1), title='(X,Z) in microns')